/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package blending;

import java.awt.CompositeContext;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;

/**
 * The BlendingContext is an abstract class that defines the
 * encapsulated and optimized environment for a compositing operation.
 * CompositeContext objects maintain state for compositing operations.
 * In a multi-threaded environment, several contexts can exist
 * simultaneously for a single Composite object
 *
 */
public abstract class BlendingContext implements CompositeContext {

	public void compose(Raster src, Raster dstIn, WritableRaster dstOut) {

		//check if both images are compatible
		checkImageErrors(src,dstIn,dstOut);

		//compute the minimum size to apply the effect
		int width = Math.min(src.getWidth(), dstIn.getWidth());
		int height = Math.min(src.getHeight(), dstIn.getHeight());

		int[] srcPixel = new int[4];
		int[] dstPixel = new int[4];
		int[] srcPixels = new int[width];
		int[] dstPixels = new int[width];

		for (int y = 0; y < height; y++) {
			src.getDataElements(0, y, width, 1, srcPixels);
			dstIn.getDataElements(0, y, width, 1, dstPixels);
			for (int x = 0; x < width; x++) {
				// pixels are stored as INT_ARGB
				// our arrays are [R, G, B, A]
				int pixel = srcPixels[x];
				srcPixel[0] = (pixel >> 16) & 0xFF;
				srcPixel[1] = (pixel >>  8) & 0xFF;
				srcPixel[2] = (pixel      ) & 0xFF;
//				srcPixel[3] = (pixel >> 24) & 0xFF;

				pixel = dstPixels[x];
				dstPixel[0] = (pixel >> 16) & 0xFF;
				dstPixel[1] = (pixel >>  8) & 0xFF;
				dstPixel[2] = (pixel      ) & 0xFF;
				dstPixel[3] = (pixel >> 24) & 0xFF;

				if (dstPixel[3] != 0) { //apply effect only to the visible pixels
					int[] result = applyBlending(srcPixel, dstPixel);
					dstPixels[x] = (result[3] & 0xFF) << 24 | (result[0] & 0xFF) << 16 | (result[1] & 0xFF) <<  8 | result[2] & 0xFF;
				}
			}
			dstOut.setDataElements(0, y, width, 1, dstPixels);
		}
	}

	public abstract int[] applyBlending(int[] srcPixel, int[] dstPixel);

	/**
	 * check if the both images has the
	 * pixels in INT format
	 * @param src
	 * @param dstIn
	 * @param dstOut
	 */
	private void checkImageErrors(Raster src, Raster dstIn, WritableRaster dstOut) {
		if (src.getSampleModel().getDataType() != DataBuffer.TYPE_INT ||
				dstIn.getSampleModel().getDataType() != DataBuffer.TYPE_INT ||
				dstOut.getSampleModel().getDataType() != DataBuffer.TYPE_INT) {
			//TODO translation?
			throw new IllegalStateException("Source and destination must store pixels as INT.");
		}
	}

	public void dispose() {
		// TODO Auto-generated method stub

	}
}