/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package core;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Panel;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;

import javax.swing.JRootPane;

import org.eclipse.swt.SWT;
import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.ScrollBar;

import tabs.TabsContainer;

/**
 * Class that creates an SWT composite, with scroll bars,
 * that contains SWING/AWT drawing
 *
 */
public class BridgeAWT_SWT extends Composite {

	/**
	 * The SWING/AWT drawing is
	 * all done by this class
	 */
	private class DrawingContainer extends Component {

		/**
		 * Serial version UID for this class 
		 */
		private static final long serialVersionUID = -8756322301639006906L;

		/**
		 * The color of 1px border around the image area
		 */
		private final Color BORDER_COLOR = Color.BLUE;

		/**
		 * Every time a paint request is done
		 * this method paints the image and the border
		 */
		public void paint(Graphics g) {
			Graphics2D g2d = paintImage(g);
			//draw the rectangle around the image
			g2d.setColor(BORDER_COLOR);
			g2d.drawRect(0, 0, getWidth() - 1, getHeight() - 1);
		}

		/**
		 * Paint the various image layers using the given graphics
		 * @param g the given graphics used to draw the image 
		 * @return the Graphics2D representation of the given Graphics
		 */
		private Graphics2D paintImage(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;

			//draw the lower layers image
			g2d.drawImage(projectTabs.getLowerLayers(), 0, 0, null);

			//pass the g2d so the selected tab will draw the currentLayer image
			projectTabs.drawCurrentLayer(g2d);

			//draw the upper layers image
			g2d.drawImage(projectTabs.getUpperLayers(), 0, 0, null);

			//return the Graphics2D for further use
			return g2d;
		}

		public void dispose() {
			//TODO free resources
		}
	}


	private final int awtContainerCompositeHeight;

	private final int awtContainerCompositeWidth;

	/**
	 * The composite that contains the scroll bars
	 */
	private Composite scrollComposite = null;

	/**
	 * The composite that contains the SWING/AWT containers 
	 */
	private Composite awtContainerComposite = null;

	/**
	 * The SWING/AWT composite used to do all the drawing
	 */
	private DrawingContainer drawingContainer = null;

	/**
	 * The scrollcomposite's horizontal scroll bar
	 */
	private ScrollBar hBar = null;

	/**
	 * The scrollcomposite's vertical scroll bar
	 */
	private ScrollBar vBar = null;

	/**
	 * The tabs in use by this project,
	 * needed to call for drawing the various image layers
	 */
	private TabsContainer projectTabs = null;

	/**
	 * Constructs a new instance of this class given its parent,
	 * a style value describing its behavior and appearance,
	 * the tabs that are in use by the parent ProjectWindow,
	 * the desired width and height for the SWING/AWT composite
	 * 
	 * <p>
	 * The style value is either one of the style constants defined in
	 * class <code>SWT</code> which is applicable to instances of this
	 * class, or must be built by <em>bitwise OR</em>'ing together 
	 * (that is, using the <code>int</code> "|" operator) two or more
	 * of those <code>SWT</code> style constants.
	 * Style bits are also inherited from superclasses.
	 * </p>
	 *
	 * @param parent the parent project for this new instance (cannot be null)
	 * @param style the style of widget to construct
	 * @param projectTabs the tabs that are in use by the parent ProjectWindow
	 * @param awtContainerCompositeWidth desired SWING/AWT composite width
	 * @param awtContainerCompositeHeight desired SWING/AWT composite height
	 */
	public BridgeAWT_SWT(ProjectWindow parent, int style, TabsContainer projectTabs, int awtContainerCompositeWidth, int awtContainerCompositeHeight) {
		super(parent.getSShell(), style);
		this.projectTabs  = projectTabs;
		this.awtContainerCompositeWidth = awtContainerCompositeWidth;
		this.awtContainerCompositeHeight = awtContainerCompositeHeight;
		createScrollComposite();
		setLayout(new FillLayout());
	}

	/**
	 * Method that creates the scrollComposite and add's hBar and vBar listeners
	 */
	private void createScrollComposite() {

		scrollComposite = new Composite(this, SWT.H_SCROLL | SWT.V_SCROLL);
		scrollComposite.setLayout(null);

		createAwtContainerComposite();

		hBar = scrollComposite.getHorizontalBar();
		vBar = scrollComposite.getVerticalBar();

		hBar.addListener (SWT.Selection, new Listener () {
			public void handleEvent (Event e) {
				processHBar();
			}
		});
		vBar.addListener (SWT.Selection, new Listener () {
			public void handleEvent (Event e) {
				processVBar();
			}
		});

		hBar.setMaximum(awtContainerComposite.getBounds().width);
		vBar.setMaximum(awtContainerComposite.getBounds().height);

		scrollComposite.addControlListener(new org.eclipse.swt.events.ControlAdapter() {
			public void controlResized(org.eclipse.swt.events.ControlEvent e) {
				int xComposite = updateHBarThumb();
				int yComposite = updateVBarThumb();
				drawingContainer.setBounds(xComposite,yComposite,drawingContainer.getBounds().width,drawingContainer.getBounds().height);

				//center awtContainerComposite on scrollComposite horizontally and vertically
				int xOffset = 0, yOffset = 0;
				if (scrollComposite.getClientArea().width > awtContainerComposite.getBounds().width) {
					xOffset = (scrollComposite.getClientArea().width - awtContainerComposite.getBounds().width) / 2;
				}
				if (scrollComposite.getClientArea().height > awtContainerComposite.getBounds().height) {
					yOffset = (scrollComposite.getClientArea().height - awtContainerComposite.getBounds().height) / 2;
				}
				awtContainerComposite.setBounds(xOffset, yOffset, awtContainerComposite.getBounds().width, awtContainerComposite.getBounds().height);
			}
		});
	}

	/**
	 * Method that update's the hBar's thumb size and position
	 * when the scrollComposite is resized. If thumb's size is
	 * maximum then the scroll auto hide's itself
	 */
	private int updateHBarThumb() {
		//adjust hBar's thumb
		if (scrollComposite.getClientArea().width < awtContainerComposite.getBounds().width) {
			hBar.setVisible(true);
			hBar.setThumb(Math.abs(scrollComposite.getClientArea().width));
		} else {
			hBar.setThumb(hBar.getMaximum());
			hBar.setVisible(false);
		}

		//also drag the image if thumb is at the end
		if (hBar.getSelection() + hBar.getThumb() == awtContainerComposite.getBounds().width) {
			int xComposite = scrollComposite.getClientArea().width - awtContainerComposite.getBounds().width;
			xComposite = xComposite > 0?0:xComposite;
			return xComposite;
		} else {
			return drawingContainer.getBounds().x;
		}
	}

	/**
	 * Method that update's the vBar's thumb size and position
	 * when the scrollComposite is resized. If thumb's size is
	 * maximum then the scroll auto hide's itself
	 */
	private int updateVBarThumb() {
		//adjust vBar's thumb
		if (scrollComposite.getClientArea().height < awtContainerComposite.getBounds().height) {
			vBar.setVisible(true);
			vBar.setThumb(Math.abs(scrollComposite.getClientArea().height));
		} else {
			vBar.setThumb(vBar.getMaximum());
			vBar.setVisible(false);
		}

		//also drag the image if thumb is at the end
		if (vBar.getSelection() + vBar.getThumb() == awtContainerComposite.getBounds().height) {
			int yComposite = scrollComposite.getClientArea().height - awtContainerComposite.getBounds().height;
			yComposite = yComposite > 0?0:yComposite;
			return yComposite;
		} else {
			return drawingContainer.getBounds().y;
		}
	}

	/**
	 * Change the awtContainerComposite horizontal position
	 * when the hBar is scrolled
	 */
	private void processHBar() {
		ScrollBar hBar = scrollComposite.getHorizontalBar();
		int hSelection = hBar.getSelection ();
		drawingContainer.setBounds(-hSelection, drawingContainer.getBounds().y, drawingContainer.getBounds().width, drawingContainer.getBounds().height);
	}

	/**
	 * Change the awtContainerComposite vertical position
	 * when the vBar is scrolled
	 */
	private void processVBar() {
		ScrollBar vBar = scrollComposite.getVerticalBar();
		int vSelection = vBar.getSelection ();
		drawingContainer.setBounds(drawingContainer.getBounds().x, -vSelection, drawingContainer.getBounds().width, drawingContainer.getBounds().height);
	}

	/**
	 * This method initializes the SWT composite
	 * used as a parent for the SWING/AWT composite
	 */
	private void createAwtContainerComposite() {
		awtContainerComposite = new Composite(scrollComposite, SWT.EMBEDDED);
		awtContainerComposite.setLayout(new GridLayout());
		awtContainerComposite.setBounds(new Rectangle(0, 0, awtContainerCompositeWidth, awtContainerCompositeHeight));
		Frame frame = SWT_AWT.new_Frame(awtContainerComposite);
		frame.setBackground(Color.white);
		Panel panel = new Panel(new BorderLayout());
		frame.add(panel);
		JRootPane root = new JRootPane();
		panel.add(root);
		final java.awt.Container contentPane = root.getContentPane();
		contentPane.setLayout(new BorderLayout());
		drawingContainer = new DrawingContainer();
		contentPane.add(drawingContainer, BorderLayout.CENTER);

		//TODO at the moment the project's window is not resizable, unused code here.
		//still this code reported some crashes when middle mouse is clicked with other mouse buttons
//		contentPane.addMouseListener(new MouseAdapter() { //mouse listeners for swing/awt

//		//fires when the mouse enters the drawing area, change the cursor icon
//		public void mouseEntered(final MouseEvent evt) {
//		if (pressedButton == 2) {
//		contentPane.setCursor (new Cursor (Cursor.MOVE_CURSOR));
//		}
//		}

//		//fires when the mouse exit's the drawing area, restore the original cursor icon
//		public void mouseExited(final MouseEvent evt) {
//		//TODO necessary this?
//		contentPane.setCursor (new Cursor (Cursor.DEFAULT_CURSOR));
//		}

//		//when mouse is pressed initialize the x,y coordinates and change cursor's icon
//		public void mousePressed(final MouseEvent evt) {
//		pressedButton = evt.getButton();
//		if (evt.getButton() == 2) {
//		Display dis = awtContainerComposite.getDisplay();
//		dis.syncExec(new Runnable(){ // important to wrap Swing Event !!
//		public void run() {
//		startX = evt.getX();
//		startY = evt.getY();
//		oldX = hBar.getSelection();
//		oldY = vBar.getSelection();
//		}
//		});
//		contentPane.setCursor (new Cursor (Cursor.MOVE_CURSOR));
//		}
//		}

//		//fires when the mouse is released, restores drawin area's default cursor icon
//		public void mouseReleased(final MouseEvent evt) {
//		// TODO on mouse released the cursor should be the default tab's cursor 
//		contentPane.setCursor (new Cursor (Cursor.DEFAULT_CURSOR));
//		pressedButton = -1;
//		}
//		});

//		contentPane.addMouseMotionListener(new MouseMotionAdapter() { //mouse motion listener

//		//fires when the image is dragged in the drawing area
//		public void mouseDragged(final MouseEvent evt) {

//		if (pressedButton == 2) {
//		Display dis = awtContainerComposite.getDisplay();
//		dis.syncExec(new Runnable(){
//		public void run() {
//		hBar.setSelection(oldX + (-(evt.getX() - startX)));
//		processHBar();
//		vBar.setSelection(oldY + (-(evt.getY() - startY)));
//		processVBar();
//		}
//		});
//		}
//		}
//		});

//		contentPane.addMouseWheelListener(new MouseWheelListener() {

//		public void mouseWheelMoved(final MouseWheelEvent e) {
//		// TODO fix left/right mouse weel cliks
//		Display dis = awtContainerComposite.getDisplay();
//		dis.syncExec(new Runnable(){
//		public void run() {
//		int MULTIPLYER = 5;
//		vBar.setSelection(vBar.getSelection() + vBar.getPageIncrement() * e.getWheelRotation() * MULTIPLYER);
//		processVBar();
//		}
//		});
//		}
//		});
	}

	/**
	 * Ask the project's drawing container to repaint it's image
	 */
	public void redrawDrawingContainer() {
		drawingContainer.repaint();
	}

	@Override
	public void dispose() {
		super.dispose();
		drawingContainer.dispose();
	}

	/**
	 * Return a copy of the current drawing container's  image
	 * @return BufferedImage with the current drawn image
	 */
	public BufferedImage getCurrentImage() {
		BufferedImage result = new BufferedImage(awtContainerCompositeWidth, awtContainerCompositeHeight, BufferedImage.TYPE_INT_RGB);
		drawingContainer.paintImage(result.getGraphics());
		return result;
	}

	/**
	 * Adds the specified mouse listener to receive mouse motion events
	 * from the project's drawing container. If listener is <code>null</code>,
	 * no exception is thrown and no action is performed.
	 * @param mouseListener the mouse listener
	 */
	public void registerMouseListener(MouseListener mouseListener) {
		drawingContainer.addMouseListener(mouseListener);
	}

	/**
	 * Adds the specified mouse motion listener to receive mouse motion events
	 * from the project's drawing container. If listener is <code>null</code>,
	 * no exception is thrown and no action is performed.
	 * @param mouseMotionListener the mouse motion listener
	 */
	public void registerMouseMotionListener(MouseMotionListener mouseMotionListener) {
		drawingContainer.addMouseMotionListener(mouseMotionListener);
	}
}