/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package dialogs;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Panel;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.GeneralPath;
import java.awt.image.BufferedImage;
import java.util.ResourceBundle;

import javax.swing.JRootPane;


import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.graphics.Font;

import core.Loader;
import core.SubImage;

import utils.StockIcon;

/**
 * Instances of this class allow the user to trace
 * a path for the lips shape using a new dialog
 */
public class LipsTracer {

	/**
	 * Class that does all the SWING/AWT drawing
	 */
	private class DrawingContainer extends Component {

		/**
		 * Every class that extends Component should have a serial version
		 */
		private static final long serialVersionUID = -7944652255908169227L;		

		/**
		 * The stroke used for the drawing
		 */
		private final Stroke basicStroke = new BasicStroke(3,1,1);

		/**
		 * The stroke's color
		 */
		private final Color strokeColor = Color.RED;

		/**
		 * Alpha composite used to fill the path with transparent color
		 */
		private final AlphaComposite fillAlphaComposite;

		/**
		 * The geometric path of the lips
		 */
		private GeneralPath gpLips = null;

		/**
		 * the background image for the SWING/AWT composite
		 */
		private BufferedImage backgroundImage;

		/**
		 * Constructs a new instance of this class given its background image.
		 * @param backgroundImage the image that will be drawn into the background
		 */
		public DrawingContainer (BufferedImage backgroundImage) {

			//set the background image
			this.backgroundImage = backgroundImage;

			//initialize alpha composite and gradients
			fillAlphaComposite = AlphaComposite.getInstance(AlphaComposite.SRC_OVER,0.3f);

			//TODO check for 1st mouse button in all listeners?
			addMouseListener(new MouseAdapter() {
				public void mouseEntered(final MouseEvent evt) {
					//when the mouse enters the drawing area, change the cursor icon
					setCursor (new Cursor (Cursor.HAND_CURSOR));
				}
				public void mouseReleased(final MouseEvent evt) {
					//on mouse release close the path
					gpLips.closePath();

					//a path has been created, enable the next/finish button
					Display dis = compositeAWT_SWT.getDisplay();
					dis.syncExec(new Runnable() {
						public void run() {
							buttonFinish.setEnabled(true);
						}
					});
					repaint(); //repaint the drawing area to make changes visible
				}

				public void mousePressed(final MouseEvent evt) {
					//when mouse is pressed create a new path and register the starting point
					gpLips = new GeneralPath();
					gpLips.moveTo(evt.getX(), evt.getY());
				}
			});

			addMouseMotionListener(new MouseMotionAdapter() {
				public void mouseDragged(final MouseEvent evt) {
					//when the mouse is dragged update the path with the new line
					gpLips.lineTo(evt.getX(), evt.getY());
					repaint(); //repaint the drawing area to make changes visible
				}
			});
		}

		/**
		 * all the drawing in the composite is done here
		 */
		public void paint(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;

			//draw the background image
			g2d.drawImage(backgroundImage, 0,0, null);

			//set stroke and drawing quality
			g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			g2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
			g2d.setColor(strokeColor);
			g2d.setStroke(basicStroke);

			//draw lips path
			if (gpLips != null) {
				java.awt.Composite defaultComposite = g2d.getComposite();
				g2d.draw(gpLips);
				g2d.setComposite(fillAlphaComposite);
				g2d.fill(gpLips);
				g2d.setComposite(defaultComposite);
			}
		}
	}

	//tells where the dialog should look for the example image
	private final String LIPS_IMAGE_PATH;

	/**
	 * The shell used by this dialog
	 */
	private Shell sShell = null;

	/**
	 * The SWING/AWT composite for the drawings
	 */
	private Composite compositeAWT_SWT = null;

	/**
	 * The composite for the help and tips
	 */
	private Composite compositeTips = null;

	/**
	 * The composite for the control buttons
	 */
	private Composite compositeControls = null;

	//The control buttons
	private Button buttonFinish = null;
	private Button buttonCancel = null;

	//separator label
	private Label labelSeparator = null;

	/**
	 * The composite for the example image
	 */
	private Composite compositeExample = null;

	//The two labels used to specify the tips/help
	private Label labelTitle = null;
	private Label labelText = null;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * The SWING/AWT drawing container
	 */
	private DrawingContainer dc = null;

	/**
	 * The final lips path
	 */
	private GeneralPath lipsPath = null;

	/**
	 * The lips image class
	 */
	private final SubImage subImage;

	//TODO check for the exceptions in the comment
	/**
	 * Constructs a new instance of this class given its parent
	 * and the lips image.
	 *
	 * @param parent a shell which will be the parent of the new instance
	 * @param subImage the image that contains the lips
	 * @param lipsImagePath the path to the lips image file
	 *
	 * @exception IllegalArgumentException <ul>
	 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
	 * </ul>
	 * @exception SWTException <ul>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
	 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
	 * </ul>
	 */
	public LipsTracer(Shell parent, SubImage subImage,String lipsImagePath) {

		this.subImage = subImage;
		//initialize the string messages for this dialog
		messages = Loader.getMessages();

		//create the AWT composite with the eyes image
		dc = new DrawingContainer(subImage.getSubImage());

		//set the lips image path
		LIPS_IMAGE_PATH = lipsImagePath;

		//create the dialog window
		createSShell(parent);
	}

	/**
	 * This method initializes sShell
	 * @param parent a shell which will be the parent of the new instance
	 */
	private void createSShell(Shell parent) {

		//grid data for the separator label
		GridData gridDataSeparator = new GridData();
		gridDataSeparator.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the shell
		sShell = new Shell(parent, SWT.APPLICATION_MODAL | SWT.DIALOG_TRIM);
		sShell.setText(messages.getString("LIPS_TRACER_TITLE"));
		sShell.setLayout(new GridLayout());
		createCompositeAWT_SWT(); //create the AWT composite
		createCompositeTips(); //create the tips composite

		//initialize the separator label
		labelSeparator = new Label(sShell, SWT.HORIZONTAL | SWT.SEPARATOR);
		labelSeparator.setLayoutData(gridDataSeparator);

		createCompositeControls(); //create the lower buttons composite
		sShell.pack();
	}

	/**
	 * This method initializes compositeAWT_SWT
	 */
	private void createCompositeAWT_SWT() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();

		//initialize the composite to hold the SWING/AWT frame
		compositeAWT_SWT = new Composite(sShell, SWT.BORDER | SWT.EMBEDDED);
		compositeAWT_SWT.setLayout(new GridLayout());
		compositeAWT_SWT.setLayoutData(gridDataComposite);

		//create the SWING/AWT frame and containers
		Frame frame = SWT_AWT.new_Frame(compositeAWT_SWT);
		frame.setBackground(Color.white);
		Panel panel = new Panel(new BorderLayout());
		panel.setBackground(Color.white);
		frame.add(panel);
		JRootPane root = new JRootPane();
		panel.add(root);
		java.awt.Container contentPane = root.getContentPane();
		contentPane.setBackground(Color.white);
		contentPane.setLayout(new BorderLayout());
		contentPane.add(dc, BorderLayout.CENTER);

		//resize the composite to the SWING/AWT image size
		gridDataComposite.widthHint = dc.backgroundImage.getWidth() - 2;
		gridDataComposite.heightHint = dc.backgroundImage.getHeight() - 2;
	}

	/**
	 * This method initializes compositeTips	
	 */
	private void createCompositeTips() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the label title
		GridData gridDataLabelTitle = new GridData();
		gridDataLabelTitle.horizontalIndent = 5;
		gridDataLabelTitle.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataLabelTitle.grabExcessVerticalSpace = true;

		//grid data for the label text
		GridData gridDataLabelText = new GridData();
		gridDataLabelText.verticalAlignment = org.eclipse.swt.layout.GridData.BEGINNING;
		gridDataLabelText.grabExcessVerticalSpace = true;
		gridDataLabelText.horizontalIndent = 5;

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 2;

		//initialize the tips composite
		compositeTips = new Composite(sShell, SWT.BORDER);
		compositeTips.setLayout(gridLayoutComposite);
		compositeTips.setLayoutData(gridDataComposite);
		createCompositeExample(); //create the image example composite

		//initialize the title and the text labels
		labelTitle = new Label(compositeTips, SWT.NONE);
		labelTitle.setText(messages.getString("LIPS_TRACER_FIRST_STEP_TITLE"));
		labelTitle.setLayoutData(gridDataLabelTitle);
		labelTitle.setFont(new Font(Display.getDefault(), "Sans", 14, SWT.BOLD));
		labelText = new Label(compositeTips, SWT.NONE);
		labelText.setText(messages.getString("LIPS_TRACER_FIRST_STEP_TEXT"));
		labelText.setLayoutData(gridDataLabelText);
	}

	/**
	 * This method initializes compositeControls
	 */
	private void createCompositeControls() {

		//grid data for the buttons
		GridData gridDataButtonFinish = new GridData();
		GridData gridDataButtonCancel = new GridData();
		gridDataButtonFinish.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataButtonCancel.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data and layout for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.END;
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 3;
		gridLayoutComposite.marginHeight = 0;
		gridLayoutComposite.marginWidth = 0;

		//initialize the composite
		compositeControls = new Composite(sShell, SWT.NONE);
		compositeControls.setLayout(gridLayoutComposite);
		compositeControls.setLayoutData(gridDataComposite);

		//initialize the finish button
		buttonFinish = new Button(compositeControls, SWT.NONE);
		buttonFinish.setText(messages.getString("LIPS_TRACER_FINISH_BUTTON"));
		buttonFinish.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_OK, StockIcon.SIZE_LARGE_TOOLBAR, true)));
		buttonFinish.setLayoutData(gridDataButtonFinish);
		buttonFinish.setEnabled(false); //by default is not enabled at startup

		//initialize the cancel button
		buttonCancel = new Button(compositeControls, SWT.NONE);
		buttonCancel.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CANCEL, StockIcon.SIZE_LARGE_TOOLBAR, true)));
		buttonCancel.setText(messages.getString("LIPS_TRACER_CANCEL_BUTTON"));
		buttonCancel.setLayoutData(gridDataButtonCancel);

		//adjust the width of the buttons to be the same on all of them
		gridDataButtonFinish.widthHint = Math.max(buttonFinish.computeSize(-1, -1).x, buttonCancel.computeSize(-1, -1).x);
		gridDataButtonCancel.widthHint = Math.max(buttonFinish.computeSize(-1, -1).x, buttonCancel.computeSize(-1, -1).x);

		//buttons listeners
		buttonFinish.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				//the paths to be returned as the final result
				lipsPath = subImage.getSmallPath(dc.gpLips);
				sShell.close();
			}
		});
		buttonCancel.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				sShell.close();
			}
		});
	}

	/**
	 * This method initializes compositeExample
	 */
	private void createCompositeExample() {

		//grid data and layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		GridData gridDataComposite = new GridData();
		gridDataComposite.verticalSpan = 2;
		gridDataComposite.horizontalIndent = 20;

		//initialize the composite
		compositeExample = new Composite(compositeTips, SWT.BORDER);
		compositeExample.setLayoutData(gridDataComposite);
		compositeExample.setLayout(gridLayoutComposite);
		compositeExample.setBackgroundImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + LIPS_IMAGE_PATH));

		//resize the composite to the image size
		gridDataComposite.widthHint = compositeExample.getBackgroundImage().getBounds().width - 2;
		gridDataComposite.heightHint = compositeExample.getBackgroundImage().getBounds().height- 2;
	}

	/**
	 * Makes the dialog visible and brings it to the front
	 * of the display.
	 *
	 * @return a GeneralPath, for the lips or null if
	 *         the dialog was canceled or an error occurred
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the dialog has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the dialog</li>
	 * </ul>
	 */
	public GeneralPath open() {

		//open the shell and wait to finish
		sShell.open();
		while (!sShell.isDisposed()) {
			if (!sShell.getDisplay().readAndDispatch())
				sShell.getDisplay().sleep();
		}

		//clear paths and image in AWT
		dc.gpLips = null;

		//return the final result
		return lipsPath;
	}
}