/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs;

import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Layout;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Widget;

/**
 * Instances of this class are controls that allow the user
 * to choose an item from a list of items.
 * Often, <code>ColorChooserButton</code>s are used in the same place
 * where a single selection <code>Table</code> widget could
 * be used but space is limited. A <code>ColorChooserButton</code> takes
 * less space than a <code>Table</code> widget and shows
 * similar information.
 * <p>
 * Note that although this class is a subclass of <code>Composite</code>,
 * it does not make sense to add children to it, or set a layout on it.
 * </p>
 * <dl>
 *
 * @see Table
 */
public class ColorChooserButton extends Composite {

	/**
	 * Class that creates a new shell with the table in it
	 * Every component created in this class has SWT.NO_FOCUS
	 * so that the pop up will not steal the main window focus
	 *
	 */
	private class TableList {

		/**
		 * The shell to be used
		 */
		private Shell sShell = null;

		/**
		 * Composite used to draw the outer border (needed, for some reason the shell border does not appear)
		 */
		private Composite preComposite = null;

		/**
		 * Composite used to place in the table
		 */
		private Composite composite = null;

		/**
		 * Table used for the elements
		 */
		private Table table = null;
		
		/**
		 * The computed width of the table (sum of all columns width)
		 */
		private int tableWidth = -1;

		/**
		 * Default constructor
		 * @param columnWidth
		 */
		public TableList(int[] columnWidth) {

			tableWidth = calculateTableWidth(columnWidth);

			GridLayout gridLayout = new GridLayout();
			gridLayout.marginWidth = 0;
			gridLayout.marginHeight = 0;
			sShell =  new Shell(SWT.ON_TOP | SWT.NO_FOCUS | SWT.BORDER);
			sShell.setLayout(gridLayout);
			createCompositeAndTable(columnWidth);

			//the default height when no elements in table is 0
			table.setSize(table.computeSize(tableWidth, 0));
			sShell.pack();
		}

		/**
		 * Given an array of column widths, return the final width needed by the table
		 * @param columnWidth the column widths array
		 * @return needed table width to show correctly all columns
		 */
		private int calculateTableWidth(int[] columnWidth) {
			int finalWidth = 0;
			for (int i = 0; i < columnWidth.length; i++) {
				finalWidth += columnWidth[i];
			}
			return finalWidth;
		}

		/**
		 * Create the pre-composite on shell, the composite that holds the table
		 * and the new table
		 * @param columnWidth the column widths array
		 */
		private void createCompositeAndTable(int[] columnWidth) {

			preComposite = new Composite(sShell, SWT.BORDER | SWT.NO_FOCUS);
			preComposite.setLayout(new GridLayout());
			composite = new Composite(preComposite, SWT.NO_FOCUS);

			//create table and the columns
			table = new Table(composite, SWT.V_SCROLL | SWT.BORDER | SWT.FULL_SELECTION | SWT.NO_FOCUS);
			table.setHeaderVisible(false);
			table.setLinesVisible(false);
			for (int i = 0; i < columnWidth.length; i++) {
				TableColumn tableColumn = new TableColumn(table, SWT.NONE);
				tableColumn.setWidth(columnWidth[i]);
			}

			//table listeners
			table.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
				public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
					tableItemSelected();
					showTableList();
				}
			});
			table.addMouseListener(new org.eclipse.swt.events.MouseAdapter() {
				public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
					sShell.setVisible(false);
				}
			});
			
			// windows hack (unreachable code in Linux)
			table.addFocusListener(new FocusAdapter() {
				public void focusGained(FocusEvent e) {
					button.setEnabled(false);
				}
				
				public void focusLost(FocusEvent e) {
					hideTableList();
					button.setEnabled(true);
				}
			});
		}

	}

	/**
	 * The button used by this widget
	 */
	private Button button;

	/**
	 * Pop up shell with table elements
	 */
	private TableList tableList;
	
	/**
	 * The number of items that are visible in the table, initialized with it's default value
	 */
	private int visibleItemCount = 5;

	/**
	 * The number of lines that the selection will jump when page
	 * up/down is clicked, initialized with it's default value
	 */
	private int pageIncrement = 5;

	private SelectionAdapter tableSelectionListener;

	/**
	 * Create a new instance of ColorChooserButton, given it's parent composite and the columns widths
	 * @param parent the parent composite
	 * @param columnWidth the column widths array
	 */
	public ColorChooserButton(Composite parent, int[] columnWidth) {
		super(parent,SWT.NONE); //this composite has no SWT style and locked with FillLayout
		super.setLayout(new FillLayout());

		//initialize the widget's button and pop up
		button = new Button(this, SWT.NONE);
		tableList = new TableList(columnWidth);

		//button listeners
		button.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				if (e.widget instanceof Button) {
					if (tableList.sShell.isVisible()) {
						hideTableList();
					} else {
						showTableList();
					}
				}
			}
		});

		//because the table can't have focus the button needs to notify the table that important keys were pressed
		button.addKeyListener(new org.eclipse.swt.events.KeyAdapter() {
			public void keyPressed(org.eclipse.swt.events.KeyEvent e) {
				switch (e.keyCode) {
				case 16777217: //key up
					tableList.table.select(tableList.table.getSelectionIndex() - 1);
					tableItemSelected();
					showTableList();
					break;
				case 16777218: //key down
					tableList.table.select(tableList.table.getSelectionIndex() + 1);
					tableItemSelected();
					showTableList();
					break;
				case 16777221: //page up
					tableList.table.select(Math.max((tableList.table.getSelectionIndex() - pageIncrement), 0));
					tableItemSelected();
					showTableList();
					break;
				case 16777222: //page down
					int selectIndex = Math.min((tableList.table.getSelectionIndex() + pageIncrement), tableList.table.getItemCount() - 1);
					tableList.table.select(selectIndex);
					tableItemSelected();
					showTableList();
					break;
				case 16777223: //home
					tableList.table.select(0);
					tableItemSelected();
					showTableList();
					break;
				case 16777224: //end
					tableList.table.select(tableList.table.getItemCount() - 1);
					tableItemSelected();
					showTableList();
					break;

				default:
					break;
				}

			}
		});

		//if the button looses focus the pop up should disappear because it has SWT.ON_TOP flag
		button.addFocusListener(new org.eclipse.swt.events.FocusAdapter() {
			public void focusLost(org.eclipse.swt.events.FocusEvent e) {
				if (System.getProperty("os.name").equalsIgnoreCase("linux")) {
					hideTableList();
				}
			}
		});
	}

	/**
	 * Update the button values every time a table item is selected
	 */
	private void tableItemSelected() {
		try {
			button.setImage(tableList.table.getSelection()[0].getImage(0));
			button.setToolTipText(tableList.table.getSelection()[0].getText(1));
			if (tableSelectionListener != null) { //notify image on button changed
				Event evt = new Event();
				evt.widget = button;
				evt.item = tableList.table.getSelection()[0];
				tableSelectionListener.widgetSelected(new SelectionEvent(evt));
			}
		} catch (Exception e) {
			//possible table is empty. no problem
		}
	}

	/**
	 * Show the table by making it's shell visible,
	 * recalculate table's size and very important,
	 * the shell's location
	 *
	 */
	private void showTableList() {
		//TODO the table height is a little bigger than the wanted... good or bad?
		//TODO if the tableHeight exceeds the screen height the visible item count should be forced to be reduced
		//so that the table shell does not exceed the screen top or bottom bounds

		//set the table height
		int tableHeight = tableList.table.getItemHeight() * visibleItemCount;
		if (visibleItemCount == 0 || visibleItemCount > tableList.table.getItemCount()) {
			tableHeight = tableList.table.getItemHeight() * tableList.table.getItemCount();
		}
		tableList.table.setSize(tableList.table.computeSize(tableList.tableWidth, tableHeight));

		//set the shell size
		tableList.sShell.pack();

		//the shell offset from the button top left location
		Point newShelloffset = new Point(0,button.getBounds().height);

		//verify if the shell doesen't need to be relocated because it passes the screen edges
		Rectangle screenSize = Display.getCurrent().getBounds(); //get the user's screen size
		Rectangle shellBounds = tableList.sShell.getBounds(); //get shell's current bounds
		if (button.toDisplay(0, button.getBounds().height).y + shellBounds.height > screenSize.height) { //check if passes bottom edge
			newShelloffset.y = -shellBounds.height;
		}
		if (button.toDisplay(0, button.getBounds().height).x + shellBounds.width > screenSize.width) { //check if passes right edge
			if (button.toDisplay(0, button.getBounds().height).x + button.getBounds().width > screenSize.width) {
				//button in not even all visible, no need to recalculate x position
			} else { //button is all visible, recalculate shell's x position
				newShelloffset.x = -(button.toDisplay(0, button.getBounds().height).x + shellBounds.width - screenSize.width);
			}
		}
		tableList.sShell.setLocation(button.toDisplay(newShelloffset));
		tableList.sShell.setVisible(true);
		if (System.getProperty("os.name").toLowerCase().contains("windows")) {
			tableList.table.forceFocus(); //windows hack
		}
	}

	/**
	 * hide the table by setting it's shell visible = false
	 */
	private void hideTableList() {
		tableList.sShell.setVisible(false);
	}

	//TODO check if comments are correct (especially exceptions)
	/**
	 * Sets the layout which is associated with the receiver to be
	 * the argument which may be null.
	 *
	 * @param layout the receiver's new layout or null
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	@Override
	public void setLayout(Layout layout) {
		// TODO Auto-generated method stub
	}

	/**
	 * Sets the receiver's text.
	 * <p>
	 * This method sets the button label.  The label may include
	 * the mnemonic character but must not contain line delimiters.
	 * </p>
	 * <p>
	 * Mnemonics are indicated by an '&amp;' that causes the next
	 * character to be the mnemonic.  When the user presses a
	 * key sequence that matches the mnemonic, a selection
	 * event occurs. On most platforms, the mnemonic appears
	 * underlined but may be emphasised in a platform specific
	 * manner.  The mnemonic indicator character '&amp;' can be
	 * escaped by doubling it in the string, causing a single
	 * '&amp;' to be displayed.
	 * </p>
	 * 
	 * @param string the new text
	 *
	 * @exception IllegalArgumentException <ul>
	 *    <li>ERROR_NULL_ARGUMENT - if the text is null</li>
	 * </ul>
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setText(String string) {
		button.setText(string);
	}

	/**
	 * Sets the receiver's tool tip text to the argument, which
	 * may be null indicating that no tool tip text should be shown.
	 *
	 * @param string the new tool tip text (or null)
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	@Override
	public void setToolTipText(String string) {
		button.setToolTipText(string);
	}

	/**
	 * Sets the number of items that are visible in the drop
	 * down portion of the receiver's list.
	 * A value of 0 or a negative value means all items
	 * 
	 * <p>
	 * The default visible item count value is 5
	 * </p>
	 * 
	 * @param count the new number of items to be visible
	 * 
	 */
	public void setVisibleItemCount(int count) {
		if (count < 0) {
			count = 0;
		}
		visibleItemCount = count;
	}

	//TODO check if comments are correct (especially exceptions)
	/**
	 * Constructs a new instance of TableItem class given a style value
	 * describing its behavior and appearance.
	 * The item is added to the end of the items maintained by the ColorChooserButton.
	 * <p>
	 * The style value is either one of the style constants defined in
	 * class <code>SWT</code> which is applicable to instances of this
	 * class, or must be built by <em>bitwise OR</em>'ing together 
	 * (that is, using the <code>int</code> "|" operator) two or more
	 * of those <code>SWT</code> style constants. The class description
	 * lists the style constants that are applicable to the class.
	 * Style bits are also inherited from superclasses.
	 * </p>
	 *
	 * @param style the style of item to construct
	 *
	 * @exception IllegalArgumentException <ul>
	 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
	 * </ul>
	 * @exception SWTException <ul>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
	 * </ul>
	 *
	 * @see SWT
	 * @see Widget#getStyle
	 * @return the instance o the created TableItem
	 */
	public TableItem addTableItem(int style) {
		return new TableItem(tableList.table, style);
	}

	//TODO check if comments are correct (especially exceptions)
	/**
	 * Sets the amount that the receiver's value will be
	 * modified by when the page increment/decrement areas
	 * are selected to the argument, which must be at least
	 * one.
	 * <p>
	 * The default page increment value is 5
	 * </p>
	 *
	 * @param pageIncrement the page increment (must be greater than zero)
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setPageIncrement(int value) {
		if (value < 1) return;
		pageIncrement = value;
	}

	/**
	 * Selects the item at the given zero-relative index in the receiver's 
	 * table list.  If the item at the index was already selected, it remains
	 * selected. Indices that are out of range are ignored.
	 *
	 * @param index the index of the item to select
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void select(int index) {
		tableList.table.select(index);
		tableItemSelected();
	}

	/**
	 * Deselects selected item in the receiver.
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void deselect() {
		tableList.table.deselectAll();
		button.setImage(null);
		button.setToolTipText(null);
	}

	/**
	 * Returns the zero-relative index of the item which is currently
	 * selected in the receiver, or -1 if no item is selected.
	 *
	 * @return the index of the selected item
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public int getSelectionIndex() {
		return tableList.table.getSelectionIndex();
	}

	public void addTableSelectionListener(org.eclipse.swt.events.SelectionAdapter listener) {
		tableSelectionListener = listener;
	}

	/**
	 * Adds the listener to the collection of listeners who will
	 * be notified when mouse buttons are pressed and released, by sending
	 * it one of the messages defined in the <code>MouseListener</code>
	 * interface.
	 *
	 * @param listener the listener which should be notified
	 *
	 * @exception IllegalArgumentException <ul>
	 *    <li>ERROR_NULL_ARGUMENT - if the listener is null</li>
	 * </ul>
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 *
	 * @see MouseListener
	 * @see #removeMouseListener
	 */
	public void addMouseListener(MouseListener listener) {
		button.addMouseListener(listener);
	}
}