/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EmptyStackException;
import java.util.Stack;

import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;

import core.ProjectWindow;
import dialogs.EyesTracer;
import dialogs.LipsTracer;

/**
 * Instances of this class that represents
 * the TabFolder to be used by the ProjectWindow
 *
 */
public class TabsContainer extends Composite {

	/**
	 * Instances of this class represents a history
	 * step for one tab or for an array of tabs
	 */
	private class HistoryStep {

		/**
		 * Save the snapshot of only one tab
		 */
		private Snapshot one = null;

		/**
		 * Save the snapshot of various tabs
		 */
		private ArrayList<Snapshot> various = null;

		/**
		 * Create an history step with the snapshot of only one tab
		 * @param one the tab's snapshot to save into history step
		 */
		public HistoryStep(Snapshot one) {
			this.one = one;
		}

		/**
		 * Create an history step with the snapshots of various tabs
		 * @param various the array of snaphots to save into history step
		 */
		public HistoryStep(ArrayList<Snapshot> various) {
			this.various = various;
		}

	}

	/**
	 * Instances of this class are used to create a new snapshot
	 * of the given tab by getting it's current state so that it
	 * can be saved in one history step
	 */
	private class Snapshot {

		/**
		 * The tab that we are going to save
		 */
		private TabContainer tab;

		/**
		 * The current state of the tab
		 */
		private Object tabState;

		/**
		 * Create a snapshot of the given tab's current state
		 * @param tab the tab for which to create the snapshot
		 * 
		 */
		public Snapshot(TabContainer tab) {
			this.tab = tab;
			this.tabState = tab.getCurrentState();
		}
	}

	/**
	 * Instances of this class represent a stack that
	 * has a maximum size; if maximum size is reached
	 * the first element will be removed
	 *
	 * @param <E> the type of the element
	 */
	private class UndoStack<E> {

		private ArrayList<E> list = null;
		private int maxSize;

		/**
		 * 
		 * @param maxSize
		 */
		//@requires maxSize > 0
		public UndoStack(int maxSize) {
			list = new ArrayList<E>();
			this.maxSize = maxSize;
		}

		/**
		 * Removes all of the elements from this stack.
		 * The stack will be empty after this call returns.
		 */
		public void clear() {
			list.clear();
		}

		/**
		 * Looks at the object at the top of this stack without removing it from the stack.
		 * @return the object at the top of this stack (the last item of the Vector object).
		 * @throws EmptyStackException - if this stack is empty.
		 */
		public E peek() {
			if (list.size() == 0) {
				throw new EmptyStackException();
			} else {
				return list.get(list.size() - 1);
			}
		}

		/**
		 * Removes the object at the top of this stack and returns that object as the value of this function.
		 * @return the object at the top of this stack (the last item of the Vector object).
		 * @throws EmptyStackException - if this stack is empty.
		 */
		public E pop() {
			if (list.size() == 0) {
				throw new EmptyStackException();
			} else {
				E e = list.get(list.size() - 1);
				list.remove(list.size() - 1);
				return e;
			}
		}

		/**
		 * Pushes an item onto the top of this stack.
		 * @param e the item to be pushed onto this stack.
		 */
		public void push(E e) {
			if (list.size() == maxSize) {
				list.remove(0);
			}
			list.add(e);
		}

		/**
		 * Returns the number of elements in this list.
		 * @return the number of elements in this list.
		 */
		public int size() {
			return list.size();
		}
	}

	/**
	 * The tab folder that is to be used by TabsContainer
	 */
	private TabFolder tabFolder = null;

	/**
	 * The current selected tab
	 */
	private TabContainer currnetTab = null;

	/**
	 * BufferedImage that contains all the layers, including the original image,
	 * that are situated below the selected tab layer
	 */
	private BufferedImage upperLayers = null;

	/**
	 * BufferedImage that contains all the layers
	 * that are situated over the selected tab layer
	 */
	private BufferedImage lowerLayers = null;

	/**
	 * The image that was opened for this project
	 */
	private BufferedImage originalImage = null;

	/**
	 * The width and height of the image that will be
	 * initialized by the ProjectWindow
	 */
	protected final int RESIZED_IMAGE_WIDTH;
	protected final int RESIZED_IMAGE_HEIGHT;
	protected final int MAX_UNDO_HISTORY;

	/**
	 * The list of tabs that are being used by TabFolder
	 */
	private ArrayList<TabContainer> tabsList = null;

	/**
	 * The undo and redo stack used to implement the history feature
	 */
	private UndoStack<HistoryStep> undoStack = null;
	private Stack<HistoryStep> redoStack = null;

	/**
	 * The TabsContainer's parent
	 */
	private ProjectWindow parent = null;

	/**
	 * Constructs a new instance of this class
	 * that initializes all the tabs
	 * @param parent The parent composite
	 * @param style SWT styles
	 * @param resizedImageWidth the width of the project's original image
	 * @param resizedImageHeight the height of the project's original image
	 */
	public TabsContainer(ProjectWindow parent, int style, int resizedImageWidth, int resizedImageHeight) {
		super(parent.getParentMainContainer(), style);
		this.parent = parent;
		RESIZED_IMAGE_WIDTH = resizedImageWidth;
		RESIZED_IMAGE_HEIGHT = resizedImageHeight;
		MAX_UNDO_HISTORY = parent.MAX_UNDO_HISTORY;
		initialize();
	}

	/**
	 * Method that sets the TabsContainer layout and size
	 * and call for the createTabFolder to add all tabs
	 */
	private void initialize() {

		GridLayout gridLayout = new GridLayout();
		gridLayout.marginHeight = 0;
		gridLayout.marginWidth = 0;
		this.setLayout(gridLayout);
		undoStack = new UndoStack<HistoryStep>(MAX_UNDO_HISTORY);
		redoStack = new Stack<HistoryStep>();

	}

	/**
	 * This method initializes tabFolder,
	 * creates and adds all tabs on tabFolder
	 */
	public void createTabFolder() {

		if (tabFolder != null) return; //createTabFolder can be called only once

		//layout and style for the TabFolder
		GridData gridData = new GridData();
//		int widthHint = parent.getParentMainContainer().getClientArea().width - 4;
//		int heightHint = parent.getParentMainContainer().getClientArea().height - 46;
		final int widthHint = parent.getParentMainContainer().getClientArea().width;
		final int heightHint = parent.getParentMainContainer().getClientArea().height;
		gridData.heightHint = heightHint;
		gridData.widthHint = widthHint;
		tabFolder = new TabFolder(this, SWT.NONE);
		tabFolder.setLayoutData(gridData);
		tabFolder.addListener(SWT.Resize, new Listener() {
			public void handleEvent(Event event) {
				tabFolder.setSize(widthHint, heightHint);
			}
		});
		
		registerMouseListeners();

		//initialize the list of tabs that are going to be added on TabFolder
		tabsList = new ArrayList<TabContainer>();

//		//add the hair tab on the TabFolder (higest position)
//		//TODO change 100 value
		HairTab hairTab = new HairTab(this,SWT.NONE, 100);
		TabItem tabItem = new TabItem(tabFolder, SWT.NONE);
		tabItem.setControl(hairTab);
		tabItem.setImage(hairTab.getTabImage());
		tabItem.setToolTipText(hairTab.getTabToolTipText());
		tabsList.add(hairTab);

		//add the concealer tab on the TabFolder (lowest layer)
		ConcealerTab concealerTab = new ConcealerTab(this,SWT.NONE, 1);
		TabItem tabItem1 = new TabItem(tabFolder, SWT.NONE);
		tabItem1.setControl(concealerTab);
		tabItem1.setImage(concealerTab.getTabImage());
		tabItem1.setToolTipText(concealerTab.getTabToolTipText());
		tabsList.add(concealerTab);

		//add the shadow tab on the TabFolder
		ShadowTab shadowTab = new ShadowTab(this,SWT.NONE, 2);
		TabItem tabItem2 = new TabItem(tabFolder, SWT.NONE);
		tabItem2.setControl(shadowTab);
		tabItem2.setImage(shadowTab.getTabImage());
		tabItem2.setToolTipText(shadowTab.getTabToolTipText());
		tabsList.add(shadowTab);

		//add the blush tab on the TabFolder
		BlushTab blushTab = new BlushTab(this,SWT.NONE, 3);
		TabItem tabItem3 = new TabItem(tabFolder, SWT.NONE);
		tabItem3.setControl(blushTab);
		tabItem3.setImage(blushTab.getTabImage());
		tabItem3.setToolTipText(blushTab.getTabToolTipText());
		tabsList.add(blushTab);

		//add the lips tab on the TabFolder
		LipsTab lipsTab = new LipsTab(this,SWT.NONE, 4);
		TabItem tabItem4 = new TabItem(tabFolder, SWT.NONE);
		tabItem4.setControl(lipsTab);
		tabItem4.setImage(lipsTab.getTabImage());
		tabItem4.setToolTipText(lipsTab.getTabToolTipText());
		tabsList.add(lipsTab);

		//add the lens tab on the TabFolder
		LensTab lensTab = new LensTab(this,SWT.NONE, 5);
		TabItem tabItem5 = new TabItem(tabFolder, SWT.NONE);
		tabItem5.setControl(lensTab);
		tabItem5.setImage(lensTab.getTabImage());
		tabItem5.setToolTipText(lensTab.getTabToolTipText());
		tabsList.add(lensTab);

		//add the free draw tab on the TabFolder
		FreeDrawTab freeDrawTab = new FreeDrawTab(this,SWT.NONE, 6);
		TabItem tabItem6 = new TabItem(tabFolder, SWT.NONE);
		tabItem6.setControl(freeDrawTab);
		tabItem6.setImage(freeDrawTab.getTabImage());
		tabItem6.setToolTipText(freeDrawTab.getTabToolTipText());
		tabsList.add(freeDrawTab);

		tabFolder.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				if (!(e.widget instanceof TabFolder)) return ;

				Control selectedTab = tabFolder.getItem(tabFolder.getSelectionIndex()).getControl();

				//disable all tabs adjustments in the project's menu
				parent.setEyesTracerEnabled(false);
				parent.setLipsTracerEnabled(false);
				parent.setLensTracerEnabled(false);

				//adjust the show hair option in the project's menu
				if (selectedTab instanceof HairTab) {
					//hair is always visible when in hair tab
					parent.setShowHairSelected(true);
					parent.setShowHairEnabled(false);
				} else {
					parent.setShowHairEnabled(true);
					resplitLayers();
				}

				//enable only the required adjustment option in the project's menu
				if (selectedTab instanceof ShadowTab) {
					parent.setEyesTracerEnabled(true);
				} else if (selectedTab instanceof LipsTab) {
					parent.setLipsTracerEnabled(true);
				} else if (selectedTab instanceof LensTab) {
					parent.setLensTracerEnabled(true);
				}
			}
		});

		//by default select the first tab
		tabFolder.setSelection(0);
		Event e = new Event();
		e.widget = tabFolder;
		tabFolder.notifyListeners(SWT.Selection, e);

		//sort the tabs by their layer position
		Collections.sort(tabsList);

		//the method resplitLayers must be called after all tabs were created and initialized on the TabFolder
		resplitLayers();
	}

	private void registerMouseListeners() {
		MouseListener mouseListener = new MouseListener() {

			public void mouseClicked(MouseEvent e) {
				currnetTab.mouseListener().mouseClicked(e);
			}

			public void mouseEntered(MouseEvent e) {
				currnetTab.mouseListener().mouseEntered(e);
			}

			public void mouseExited(MouseEvent e) {
				currnetTab.mouseListener().mouseExited(e);
			}

			public void mousePressed(MouseEvent e) {
				currnetTab.mouseListener().mousePressed(e);
			}

			public void mouseReleased(MouseEvent e) {
				currnetTab.mouseListener().mouseReleased(e);
			}

		};

		MouseMotionListener mouseMotionListener = new MouseMotionListener() {

			public void mouseDragged(MouseEvent e) {
				currnetTab.mouseMotionListener().mouseDragged(e);
			}

			public void mouseMoved(MouseEvent e) {
				currnetTab.mouseMotionListener().mouseMoved(e);
			}

		};

		parent.registerMouseListener(mouseListener);
		parent.registerMouseMotionListener(mouseMotionListener);
	}

	/**
	 * Method that splits all the tab layers in three:
	 * -lowerLayers: all the layers including the original image that are below the selected tab layer
	 * -upperLayers: all the layers that are above the selected tab layer
	 * -currentLayer: the layer of the selected tab
	 * This is used so that the DrawingContainer will only draw 3 images every time a change is made
	 * to selected tab's layer (lower,current,upper layers) instead of drawing every layer one by one
	 */
	private void resplitLayers() {

		//set's the new selected tab available to be used by getCurrentTab
		currnetTab = (TabContainer) tabFolder.getItem(tabFolder.getSelectionIndex()).getControl();

		//currentTab is being used by lower and upper methods so needs to be initialized before calling the methods
		resplitLayers_Lower();
		resplitLayers_upper();

	}

	/**
	 * Method that creates an empty transparent image and begins to draw
	 * the original image and all the layers below the selected tab layer by order
	 */
	private void resplitLayers_Lower() {

		lowerLayers = new BufferedImage(RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2d = lowerLayers.createGraphics();
		if (originalImage == null) { //draw white background if no original image was set yet
			g2d.setColor(Color.white);
			g2d.fillRect(0,0, RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT);
		}
		g2d.drawImage(originalImage, 0, 0, null);
		for (TabContainer t : tabsList) {
			if (t.getTabLayerPosition() < currnetTab.getTabLayerPosition()) {
				g2d.drawImage(t.getLayerImage(), 0,0, null);
			} else {
				break;
			}
		}
		g2d.dispose();

	}

	/**
	 * Method that creates an empty transparent image and begins to draw
	 * all the layers that are above the selected tab layer by order
	 */
	private void resplitLayers_upper() {

		upperLayers = new BufferedImage(RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2d = upperLayers.createGraphics();
		for (int i = tabsList.indexOf(currnetTab) + 1; i < tabsList.size(); i++) {
			TabContainer t = tabsList.get(i);
			g2d.drawImage(t.getLayerImage(), 0,0, null);
		}
		g2d.dispose();

	}

	/**
	 * Method that returns an image containing all the layers
	 * that are above the selected tab layer
	 * @return BufferedImage with all the layers above the selected tab layer
	 */
	public BufferedImage getUpperLayers() {
		return upperLayers;
	}

	/**
	 * Method that returns the selected tab layer
	 * @return BufferedImage that contains the selected tab layer
	 * 
	 */
	public BufferedImage getCurrentLayer() {
		return currnetTab.getLayerImage();
	}

	/**
	 * Calls the current tab to draw it's resulting image on the DrawingContainer
	 * using DrawingContainer's given Graphics2D
	 * @param g2d DrawingContainer's Graphics2D
	 */
	public void drawCurrentLayer(Graphics2D g2d) {
		currnetTab.drawCurrentLayer(g2d);
	}

	/**
	 * Method that returns an image containing all the layers
	 * that are below the selected tab layer, including the original image
	 * @return BufferedImage with all the layers below the selected tab layer
	 */
	public BufferedImage getLowerLayers() {
		return lowerLayers;
	}

	/**
	 * Method that returns the selected tab layer
	 * @return TabContainer with the selected tab
	 */
	public TabContainer getCurrentTab() {
		return currnetTab;
	}

	/**
	 * Method that initializes the originalImage with the image read
	 * from file by the ProjectWindow class
	 * @param originalImage
	 */
	public void setOriginalImage(BufferedImage originalImage) {
		//TODO a copy of the original image is created for use by tabs, maybe not necessary...
		//creates an blank transparent image and draws the original image on it
		int w = originalImage.getWidth();
		int h = originalImage.getHeight();
		this.originalImage = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2d = this.originalImage.createGraphics();
		g2d.drawImage(originalImage, 0, 0, null);
		g2d.dispose();

		//we need to redraw only the lower layers because the original image was changed
		resplitLayers_Lower();
	}

	//TODO add comment if there is no other way of doing this (TabContainer super parent)
	TabFolder getTabFolder() {
		return tabFolder;
	}

	/**
	 * Method that goes one step back in the changes history
	 */
	public void undo() {

		//get the history step from undo stack		
		HistoryStep fromUndo = undoStack.pop();

		if (fromUndo.one != null) { //check if only one tab changed during the change in this step

			//create a new history step taking a snapshot of the tab's current state and place it to redo stack
			HistoryStep toRedo = new HistoryStep(new Snapshot(fromUndo.one.tab));
			redoStack.push(toRedo);

			//set the current state of the tab to be the state taken from undo stack
			fromUndo.one.tab.setCurrentState(fromUndo.one.tabState);

		} else { //if various tabs were changed during the change in this step

			//a list to save the current snapshots of all tabs
			ArrayList<Snapshot> list = new ArrayList<Snapshot>();
			for (TabContainer t : tabsList) {
				list.add(new Snapshot(t));
			}

			//create a history step with current tabs state and place it into redo stack
			HistoryStep toRedo = new HistoryStep(list);
			redoStack.push(toRedo);

			//for each tab set the current state to be the state taken from undo stack
			for (Snapshot c : fromUndo.various) {
				c.tab.setCurrentState(c.tabState);
			}

		}

		//adjust the undo and redo buttons
		parent.setEnabledRedo(true);
		if (undoStack.size() == 0) {
			parent.setEnabledUndo(false);
		}

		//redraw all the layers so that the changes take effect
		resplitLayers();

	}

	/**
	 * Method that goes one step forward in the changes history
	 */
	public void redo() {

		//get the history step from the redo stack
		HistoryStep fromRedo = redoStack.pop();

		if (fromRedo.one != null) { //check if only one tab changed during the change in this step

			//create a new history step taking a snapshot of the tab's current state and place it to undo stack
			HistoryStep toUndo = new HistoryStep(new Snapshot(fromRedo.one.tab));
			undoStack.push(toUndo);

			//set the current state of the tab to be the state taken from redo stack
			fromRedo.one.tab.setCurrentState(fromRedo.one.tabState);

		} else { //if various tabs were changed during the change in this step

			//a list to save the current snapshots of all tabs
			ArrayList<Snapshot> list = new ArrayList<Snapshot>();
			for (TabContainer t : tabsList) {
				list.add(new Snapshot(t));
			}

			//create a history step with current tabs state and place it into undo stack
			HistoryStep toUndo = new HistoryStep(list);
			undoStack.push(toUndo);

			//for each tab set the current state to be the state taken from redo stack
			for (Snapshot c : fromRedo.various) {
				c.tab.setCurrentState(c.tabState);
			}
		}

		//adjust the undo and redo buttons
		parent.setEnabledUndo(true);
		if (redoStack.size() == 0) {
			parent.setEnabledRedo(false);
		}

		//redraw all the layers so that the changes take effect
		resplitLayers();

	}

	/**
	 * Method that resets all the changes made to the original's project image 
	 */
	public void clear() {

		//a list to save the current snapshots of all tabs and calls each tab to clear it's changes to the image
		ArrayList<Snapshot> list = new ArrayList<Snapshot>();
		for (TabContainer t : tabsList) {
			list.add(new Snapshot(t));
			t.clear();
		}

		//create a history step with current tabs state and place it into undo stack
		HistoryStep toUndo = new HistoryStep(list);
		undoStack.push(toUndo);

		//adjust the undo and redo buttons
		parent.setEnabledUndo(true);
		parent.setEnabledRedo(false);
		redoStack.clear();

		//redraw all the layers so that changes take effect
		resplitLayers();

	}

	/**
	 * Call for a complete redrawing of the project's drawing area
	 * <p>
	 * Note that only the current tab has access to this feature.
	 * Other tabs calls that are in the background will be ignored
	 * </p>
	 * @param callerTab the tab that calls for the redrawing
	 */
	public void redrawImageArea(TabContainer callerTab) {
		//only selected tab has access to call for a redrawing
		if (callerTab.equals(currnetTab)) {
			parent.redrawImageArea();
		}
	}

	/**
	 * Create a history step for the current active tab
	 */
	public void createHistoryStep() {
		//create a new history step taking a snapshot of the tab's current state and place it to undo stack 
		int usedTabIndex = tabsList.indexOf(currnetTab);
		HistoryStep toUndo = new HistoryStep(new Snapshot(tabsList.get(usedTabIndex)));
		undoStack.push(toUndo);

		//adjust the undo and redo buttons and clear the redo stack
		parent.setEnabledUndo(true);
		parent.setEnabledRedo(false);
		redoStack.clear();

	}

	/**
	 * Returns the tabs container associated project
	 * @return the tabs container associated project
	 */
	public ProjectWindow getParentProject() {
		return parent;
	}

	/**
	 * Returns a copy of the project's original image
	 * @return a copy of the project's original image
	 */
	public BufferedImage getOriginalImage() {
		return originalImage.getSubimage(0, 0, originalImage.getWidth(), originalImage.getHeight());
	}

	/**
	 * Opens a dialog that allows the user to trace the eyes path for the shadow tab
	 */
	public void setShadowPath() {
		for (int i = 0; i < tabFolder.getItemCount(); i++) {
			if (tabFolder.getItem(i).getControl() instanceof ShadowTab) {
				ShadowTab shadow = (ShadowTab) tabFolder.getItem(i).getControl();
				shadow.setEyesPath(new EyesTracer(parent.getSShell(), parent.eyesSubImage, "eye-left.png", "eye-right.png").open());
				break;
			}
		}
	}

	/**
	 * Opens a dialog that allows the user to trace the eyes path for the lens tab
	 */
	public void setLensPath() {
		for (int i = 0; i < tabFolder.getItemCount(); i++) {
			if (tabFolder.getItem(i).getControl() instanceof LensTab) {
				LensTab lens = (LensTab) tabFolder.getItem(i).getControl();
				lens.setEyesPath(new EyesTracer(parent.getSShell(), parent.eyesSubImage, "lens-left.png", "lens-right.png").open());
				break;
			}
		}
	}

	/**
	 * Opens a dialog that allows the user to trace the lips path for the lips tab
	 */
	public void setLipsPath() {
		for (int i = 0; i < tabFolder.getItemCount(); i++) {
			if (tabFolder.getItem(i).getControl() instanceof LipsTab) {
				LipsTab lens = (LipsTab) tabFolder.getItem(i).getControl();
				lens.setLipsPath(new LipsTracer(parent.getSShell(), parent.lipsSubImage, "lips-full.png").open());
				break;
			}
		}
	}

	/**
	 * Hides or unhides the {@link HairTab}'s hair image
	 * @param selection true if the hair should be visible, false otherwise
	 */
	public void changeHairVisibility(boolean selection) {
		for (int i = 0; i < tabFolder.getItemCount(); i++) {
			if (tabFolder.getItem(i).getControl() instanceof HairTab) {
				HairTab hairTab = (HairTab) tabFolder.getItem(i).getControl();
				hairTab.setHairIsVisible(selection);
				break;
			}
		}
		//hair tab may be in one of the layers so we need to re-split them for changes to take effect on the original image
		resplitLayers();
	}

	/**
	 * Notify all tabs that the projec't original image was changed
	 * (size, rotation, position)
	 */
	public void originalImageChanged() {
		for (int i = 0; i < tabFolder.getItemCount(); i++) {
			TabContainer tab = (TabContainer) tabFolder.getItem(i).getControl();
			tab.originalImageChanged();
		}
	}
	
	/**
	 * Dispose all unused resources to reduce memory leaks
	 */
	public void dispose() {
		super.dispose();
		for (TabContainer t : tabsList) {
			t.dispose();
		}
		lowerLayers = null;
		originalImage = null;
		upperLayers = null;
	}
}